import calendar
import datetime
import pytz

from django.conf.urls.defaults import *
from django.utils.translation import ugettext_lazy as _
from django_vcs_watch.models import Repository, Commit, FeedItem
from django_vcs_watch.forms import RepositoryForm
from django_vcs_watch.feeds import LatestRepositories, LatestCommits

from django_globals import globals

def timestamp():
    return calendar.timegm(pytz.utc.localize(datetime.datetime.utcnow()).timetuple())

def get_tb(request):
    if 'tb' in request.GET:
        return pytz.utc.localize(
            datetime.datetime.utcfromtimestamp(
                int(request.GET.get('tb'))))
    return pytz.utc.localize(datetime.datetime.utcnow())


repository_detail = {
    'cls': Repository,
    'query': lambda request, slug: {
        'slug': slug,
    },
    'template_name': 'django_vcs_watch/repository_detail.html',
}

user_commits_dict = dict(
    extra_context = dict(title = _('User commits')),
    template = 'django_vcs_watch/user_commits.html',
)

feed_items = {
    'cls': FeedItem,
    'query': lambda request, slug: {
        'slug': slug,
        'date' : {'$lt': get_tb(request)}
    },
    'paginate_by': 20,
    'template_name': 'django_vcs_watch/repository_detail.html',
    'template_object_name': 'commits_list',
    'map_func': lambda x: x.commit,
}

feeds = {
    'commits': LatestCommits,
    'repositories': LatestRepositories,
}

main_page = {
    'template': 'django_vcs_watch/main.html',
    'extra_context': {
        'repositories': lambda: Repository.objects.filter(public=True).exclude(updated_at=None)[:10],
    }
}

autocomplete = {
    'template_name': 'django_vcs_watch/autocomplete.html'
}

create_repository_page = {
    'form_class': RepositoryForm,
    'template_name': 'django_vcs_watch/create_repository.html',
    'extra_context': { 'user': lambda: globals.user },
}

urlpatterns = patterns('django_vcs_watch.views',
   (r'^r/feed/$',
        'feed', {'slug': 'repositories', 'feed_dict': feeds}, 'vcs-watch-feed-repositories'),
   (r'^r/(?P<param>[a-z0-9-]+)/feed/$',
        'feed', {'slug': 'commits', 'feed_dict': feeds}, 'vcs-watch-feed-commits'),
   (r'^r/(?P<repository_slug>[a-z0-9-]+)/(?P<revision>[a-z0-9-]{1,36})/$',
        'commit', {}, 'vcs-watch-commit'),
   (r'^profile/$', 'profile', {}, 'vcs-watch-profile'),
   (r'^autocomplete/$', 'autocomplete', autocomplete, 'vcs-watch-autocomplete'),
   (r'^add/$', 'create', create_repository_page, 'vcs-watch-add'),
   (r'^r/(?P<slug>[a-z0-9-]+)/$',
        'object_detail', repository_detail, 'vcs-watch-repository'),
   (r'^f/(?P<slug>[a-z0-9-]+)/$',
        'object_list', feed_items, 'vcs-watch-feed'),
   (r'^refresh-feed/$',
        'refresh_feed', {}, 'vcs-watch-refresh-feed'),
   (r'^ignore/$', 'ignore', {}, 'vcs-watch-ignore'),
   (r'^unignore/$', 'unignore', {}, 'vcs-watch-unignore'),
   (r'^watch/$', 'watch', {}, 'vcs-watch-watch'),
   (r'^unwatch/$', 'unwatch', {}, 'vcs-watch-ununwatch'),
)

urlpatterns += patterns('django.views.generic',
   (r'^r/$', 'simple.direct_to_template', main_page, 'vcs-watch-repositories'),
   (r'^u/(?P<author>.*)/$', 'simple.direct_to_template', user_commits_dict, 'vcs-watch-commits-by'),
)




import uuid
from pdb import set_trace

from django import forms
from django.utils.translation import ugettext_lazy as _

from django_vcs_watch.models import Repository
from django_vcs_watch.settings import VCS_ONLY_PUBLIC_REPS, \
                                      VCS_URL_REWRITER
from django_vcs_watch.utils import make_slug



def find_unused_slug(slug):
    def slug_exists(slug):
        return len(Repository.objects.find({'slug': slug})) > 0

    if not slug_exists(slug):
        return slug

    for i in xrange(2, 1000):
        new_slug = '%s-%d' % (slug, i)
        if not slug_exists(new_slug):
            return new_slug
    raise Exception("Can't find unused slug.")



class RepositoryForm(forms.Form):
    url      = forms.CharField(max_length = 255)
    slug     = forms.SlugField(required = False)
    username = forms.CharField(max_length = 40, required = False)
    password = forms.CharField(max_length = 40, required = False, widget = forms.PasswordInput())

    def clean_url(self):
        return VCS_URL_REWRITER(self.cleaned_data.get('url', ''))

    def save(self):
        repository = Repository.objects.find_one({'url': self.cleaned_data['url']})

        data = {
            'public': True,
            'slug': None,
        }
        data.update(self.cleaned_data)

        if repository:
            if not data['slug']:
                del data['slug']
            repository.update(data)
        else:
            slug = data['slug']
            if not slug:
                slug = make_slug(data['url'])

            real_slug = find_unused_slug(slug)
            if data['slug'] != real_slug:
                data['original_slug'] = data['slug']
                data['slug'] = real_slug

            repository = Repository(**data)

        return repository.save()




import re
import md5
from datetime import datetime

from django.conf import settings
from django.utils.encoding import DjangoUnicodeDecodeError, force_unicode

# URL rewriters
_http_re = re.compile(r'^(https|http)')

def to_svn_ssh(url):
    return _http_re.sub('svn+ssh', url)

def remove_wsvn(url):
    return url.replace('/websvn/wsvn', '')

def strip_get(url):
    return url.split('?', 1)[0]

# slugificator

def make_slug(url, path_filter = lambda x: x not in ['', 'svn']):
    "Returns a slug for given URL."
    from urlparse import urlparse
    host, path = urlparse(url)[1:3]
    host = host.split('.')

    result = []

    if host[-2] == 'googlecode':
        result.append(host[-3])

    path = filter(path_filter, path.replace('.', '-').split('/'))

    def index(l, value):
        try: return l.index(value)
        except ValueError: return None

    ttb_index = filter(lambda x: x is not None, [
        index(path, 'trunk'),
        index(path, 'tags'),
        index(path, 'branches'),
    ])

    if ttb_index:
        path.pop(ttb_index[0])
    result += path

    if len(result) == 0:
        result.append(host[-2])

    return '-'.join(result)


def strip_timezone(t):
    return datetime(t.year, t.month, t.day, t.hour, t.minute, t.second, t.microsecond)


def timedelta_to_string(td):
    return '%d days, %d hours, %d minutes, %d seconds' % (
        td.days,
        td.seconds / 3600,
        td.seconds % 3600 / 60,
        td.seconds % 60
    )


def guess_encoding(s):
    options = [
        {}, {'encoding': 'cp1251'}, {'encoding': 'koi8-r'},
        {'encoding': 'utf-8', 'errors': 'ignore'}
    ]
    for o in options:
        try:
            return force_unicode(s, **o)
        except DjangoUnicodeDecodeError:
            pass
    raise Exception('Can\'t decode string: %r' % s)

def mongo():
    from pymongo.connection import Connection
    from django_vcs_watch.settings import \
        MONGO_URL, \
        MONGO_DB

    conn = Connection(MONGO_URL)
    return conn[MONGO_DB]


def get_user_feed_slug(user):
    if user.is_anonymous():
        raise Exception("Anonymous user can't have a feed.")
    return md5.md5(user.username + settings.SECRET_KEY).hexdigest()


class DiffProcessor(object):
    """Helper to parse diffs, separate by file and collect some information."""

    def _start_new_file(self, filename):
        if self.current is not None:
            self.current['diff'] = '\n'.join(self.current['diff'])
            self.changed.append(self.current)

        self.current = dict(
            filename = filename,
            diff = [],
            stats = dict(added = 0, removed = 0)
        )


    def process(self, diff):
        """Parses diff and returns dict with statistics and separate diffs for each changed file."""
        self.changed = []
        self.current = None
        self.added = 0
        self.removed = 0

        for line in diff.split('\n'):
            if line.startswith('Index: '):
                self._start_new_file(line.split(' ', 1)[1])
                continue
            elif line[:3] in ('===', '---', '+++'):
                continue
            elif line.startswith('+'):
                self.added += 1
                self.current['stats']['added'] += 1
            elif line.startswith('-'):
                self.removed -= 1
                self.current['stats']['removed'] -= 1

            if self.current is not None:
                self.current['diff'].append(line)

        self._start_new_file(None)

        return dict(
                changed = self.changed,
                stats = dict(
                        added = self.added,
                        removed = self.removed
                    )
            )




from django.test import TestCase
from django.test.client import Client
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User

from django_vcs_watch.models import Repository
from django_vcs_watch.utils import make_slug, DiffProcessor, mongo
from django_vcs_watch.settings import VCS_ONLY_PUBLIC_REPS

from pdb import set_trace

class _RepsTestCase(TestCase):
    def setUp(self):
        try:
            self.tester = User.objects.get(username='tester')
        except User.DoesNotExist:
            self.tester = User.objects.create_user(username='tester', email='tester@example.com', password='test')

        self.add_url = reverse('vcs-watch-add')
        self.login_url = reverse('vcs-watch-add')

        Repository.objects.db = mongo()
        Repository.objects.remove({})
        super(_RepsTestCase, self).setUp()

class Repositories(_RepsTestCase):
    def testGetAnonymousForm(self):
        response = self.client.get(self.add_url)
        self.assertEqual(200, response.status_code)
        self.assertContains(response, 'id_url')
        self.assertContains(response, 'id_slug')

    def testAddRepByAnonymous(self):
        url = 'http://svn.example.com/trunk/'
        response = self.client.post(self.add_url, dict(url=url))

        reps = list(Repository.objects.find())
        self.assertEqual(1, len(reps))
        self.assertEqual(url, reps[0]['url'])
        self.assertEqual(True, reps[0]['public'])

        url = reverse('vcs-watch-repository', kwargs = dict(slug=reps[0]['slug']))
        self.assertRedirects(response, url)

    def testAddRepTwice(self):
        url = 'http://svn.example.com/trunk/'
        response = self.client.post(self.add_url, dict(url=url))
        response = self.client.post(self.add_url, dict(url=url))

        reps = list(Repository.objects.all())
        self.assertEqual(1, len(reps))
        self.assertEqual(url, reps[0]['url'])
        self.assertEqual(True, reps[0]['public'])
        self.assertEqual(u'example', reps[0]['slug'])

        url = reverse('vcs-watch-repository', kwargs = dict(slug=reps[0]['slug']))
        self.assertRedirects(response, url)

if not VCS_ONLY_PUBLIC_REPS:
    class PrivateRepositories(_RepsTestCase):
        def testGetNonAnonymousForm(self):
            self.assert_(self.client.login(username='tester', password='test'))

            response = self.client.get(self.add_url)
            self.assertEqual(200, response.status_code)
            self.assertContains(response, 'id_url')
            self.assertContains(response, 'id_public')
            self.assertContains(response, 'id_username')
            self.assertContains(response, 'id_password')

        def testAddByAuthorized(self):
            self.assert_(self.client.login(username='tester', password='test'))
            url = 'http://svn.example.com/trunk/'
            response = self.client.post(self.add_url, dict(
                                            url=url,
                                            username='svnuser',
                                            password='svnpass',
                                            ))

            reps = list(Repository.objects.all())
            self.assertEqual(1, len(reps))
            self.assertEqual(url, reps[0]['url'])
            self.assertEqual(False, reps[0]['public'])
            self.assertEqual('svnuser', reps[0]['username'])
            self.assertEqual('svnpass', reps[0]['password'])

            url = reverse('vcs-watch-repository', kwargs = dict(slug=reps[0]['slug']))
            self.assertRedirects(response, url)

        def testPrivateReposList(self):
            self.assert_(self.client.login(username='tester', password='test'))

            url = 'http://svn.example.com/trunk/'
            slugs = []
            for i in xrange(10):
                response = self.client.post(self.add_url, dict(
                                                url=url,
                                                username='svnuser',
                                                password='svnpass',
                                                ))
                slugs.append(response.get('Location', None).split('/')[-2])

            user_url = reverse('vcs-watch-profile')
            response = self.client.get(user_url)
            self.assertEqual(200, response.status_code)

            for slug in slugs:
                self.assertContains(response, slug)


class Rewrites(TestCase):
    def testSSH(self):
        from django_vcs_watch.utils import to_svn_ssh

        self.assertEqual('svn+ssh://svn.example.com/test',
              to_svn_ssh('http://svn.example.com/test'))

        self.assertEqual('svn+ssh://svn.example.com/test',
              to_svn_ssh('https://svn.example.com/test'))


    def testWSVN(self):
        from django_vcs_watch.utils import remove_wsvn

        self.assertEqual('https://svn.example.org/auto',
             remove_wsvn('https://svn.example.org/websvn/wsvn/auto'))

    def testStripGET(self):
        from django_vcs_watch.utils import strip_get

        self.assertEqual('https://svn.example.org/auto',
               strip_get('https://svn.example.org/auto?'))

class Slugs(TestCase):
    def testMakeSlugs(self):
        self.assertEqual(
            'django',
            make_slug('http://svn.example.com/django/trunk/'))

        self.assertEqual(
            'example',
            make_slug('http://svn.example.com'))
        self.assertEqual(
            'example',
            make_slug('http://svn.example.com/trunk/'))
        self.assertEqual(
            'django-django-contrib',
            make_slug('http://svn.example.com/django/trunk/django/contrib/'))
        self.assertEqual(
            'django-new-admin',
            make_slug('http://svn.example.com/django/branches/new-admin/'))
        self.assertEqual(
            'django-1-1',
            make_slug('http://svn.example.com/django/tags/1.1/'))
        self.assertEqual(
            'mock',
            make_slug('http://mock.googlecode.com/svn/'))
        self.assertEqual(
            'django-grappelli-2-5',
            make_slug('http://django-grappelli.googlecode.com/svn/branches/2.5'))


class DiffProcessing(TestCase):
    def testSeparate(self):
        diff = """
Index: django/db/backends/oracle/introspection.py
===================================================================
--- django/db/backends/oracle/introspection.py  (revision 11474)
+++ django/db/backends/oracle/introspection.py  (revision 11475)
@@ -26,6 +26,14 @@
     except AttributeError:
         pass
  
+    def get_field_type(self, data_type, description):
+        # If it's a NUMBER with scale == 0, consider it an IntegerField
+        if data_type == cx_Oracle.NUMBER and description[5] == 0:
+            return 'IntegerField'
+        else:
+            return super(DatabaseIntrospection, self).get_field_type(
+                data_type, description)
+ 
     def get_table_list(self, cursor):
         \"Returns a list of table names in the current database.\"
         cursor.execute(\"SELECT TABLE_NAME FROM USER_TABLES\")
Index: django/db/backends/__init__.py
===================================================================
--- django/db/backends/__init__.py      (revision 11474)
+++ django/db/backends/__init__.py      (revision 11475)
@@ -470,6 +470,14 @@
     def __init__(self, connection):
         self.connection = connection

+    def get_field_type(self, data_type, description):
+        \"\"\"Hook for a database backend to use the cursor description to
+        match a Django field type to a database column.
+
+        For Oracle, the column data_type on its own is insufficient to
+        distinguish between a FloatField and IntegerField, for example.\"\"\"
+        return self.data_types_reverse[data_type]
+
     def table_name_converter(self, name):
         \"\"\"Apply a conversion to the name for the purposes of comparison.

@@ -560,4 +568,3 @@
     def validate_field(self, errors, opts, f):
         \"By default, there is no backend-specific validation\"
         pass
-
Index: django/core/management/commands/inspectdb.py
===================================================================
--- django/core/management/commands/inspectdb.py        (revision 11474)
+++ django/core/management/commands/inspectdb.py        (revision 11475)
@@ -73,7 +73,7 @@
                         extra_params['db_column'] = column_name
                 else:
                     try:
-                        field_type = connection.introspection.data_types_reverse[row[1]]
+                        field_type = connection.introspection.get_field_type(row[1], row)
                     except KeyError:
                         field_type = 'TextField'
                         comment_notes.append('This field type is a guess.')
"""



        processor = DiffProcessor()
        result = processor.process(diff)
        self.assertEqual(3, len(result['changed']))
        self.assertEqual([
                'django/db/backends/oracle/introspection.py',
                'django/db/backends/__init__.py',
                'django/core/management/commands/inspectdb.py',
            ], [file['filename'] for file in result['changed']])
        self.assertEqual(17, result['stats']['added'])
        self.assertEqual(-2, result['stats']['removed'])
        self.assert_(unicode, type(result['changed'][0]['diff']))

        self.assertEqual(8, result['changed'][0]['stats']['added'])
        self.assertEqual(0, result['changed'][0]['stats']['removed'])
        self.assertEqual(8, result['changed'][1]['stats']['added'])
        self.assertEqual(-1, result['changed'][1]['stats']['removed'])
        self.assertEqual(1, result['changed'][2]['stats']['added'])
        self.assertEqual(-1, result['changed'][2]['stats']['removed'])




import calendar
import datetime
import logging
import operator
import os
import pytz

from pdb import set_trace

from django.db import models
from django.conf import settings
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import DjangoUnicodeDecodeError

from django_fields.fields import EncryptedCharField

from django_vcs_watch.settings import \
    REVISION_LIMIT, \
    CHECK_INTERVAL_MIN, \
    CHECK_INTERVAL_MAX

from django_vcs_watch.utils import \
    timedelta_to_string, \
    strip_timezone, \
    mongo, \
    get_user_feed_slug

from pymongo import DESCENDING
from mongobongo import Document


if 'django_globals' not in settings.INSTALLED_APPS:
    raise Exception('Please, install django_globals application.')

if 'django_globals.middleware.User' not in settings.MIDDLEWARE_CLASSES:
    raise Exception('Please, add django_globals.middleware.User to the MIDDLEWARE_CLASSES.')


def delta2secs(d):
    return d.days * 24 * 60 * 60 + d.seconds


class Commit(Document):
    collection = 'commits'

    class Meta:
        ordering = [('date', DESCENDING)]

    def __unicode__(self):
        return _('Commit %(revision)s by %(author)s') % self.__dict__

    @models.permalink
    def get_absolute_url(self):
        return ('vcs-watch-commit', (), {'repository_slug': self.slug, 'revision': self.revision})

    def get_timestamp(self):
        if self.date is not None:
            return calendar.timegm(self.utcdate.timetuple())
        return 0

    @property
    def utcdate(self):
        return pytz.utc.localize(self._data.get('date'))



class Repository(Document):
    collection = 'repositories'

    def __unicode__(self):
        return _('Repository at %s') % self.url


    @models.permalink
    def get_absolute_url(self):
        return ('vcs-watch-repository', (), {'slug': self.slug})


    @models.permalink
    def get_rss_url(self):
        return ('vcs-watch-feed-commits', (), {'param': self.slug })


    def update(self):
        log = logging.getLogger('django_vcs_watch.repository.update')
        log.debug('updating %s' % self.slug)

        from django_vcs_watch.backends.svn import get_updates

        try:
            commits = get_updates(self.url, self.last_rev,
                                  self.username, self.password)
        except Exception, e:
            log.exception('error during repository update')
            self.last_error = str(e).decode('utf-8')

            if self.last_error_date:
                interval_to_check = min(
                    max(
                        delta2secs(datetime.datetime.utcnow() - self.last_error_date) * 2,
                        CHECK_INTERVAL_MIN
                    ),
                    CHECK_INTERVAL_MAX)
            else:
                interval_to_check = CHECK_INTERVAL_MIN

            interval_to_check = datetime.timedelta(0, interval_to_check)
            log.debug('next check will be after %s' % timedelta_to_string(interval_to_check))

            self.last_error_date = datetime.datetime.utcnow()
            self.next_check_at = self.last_error_date + interval_to_check
            self.save()
            return

        for commit in commits:
            commit = Commit(**commit)
            if len(Commit.objects.find({'slug': self.slug, 'revision': commit.revision})) > 0:
                continue

            commit.slug = self.slug

            try:
                commit.save()
            except Exception:
                # just ignore this strange errors,
                # caused by wrong encoding in the comments
                # or binary files without mime type.
                log.exception('error during commit saving %s:%s' % (self.url, commit.revision))
            self.last_rev = commit.revision
            self.updated_at = commit.date


        if len(commits) > 0:
            self.last_error = None
            self.last_error_date = None

        # don't update more often than latest commits
        # TODO remove list when mongobongo will support len on cursor proxy
        latest_commits = list(Commit.objects.find({'slug': self.slug}).sort([('date', -1)])[:3])
        assert(len(latest_commits) <= 3)

        def weight(x):
            return 1.0 / (2 ** (x+1))

        if len(latest_commits) > 0:
            deltas = [
                (weight(0),
                 (datetime.datetime.utcnow() - latest_commits[0].date))
            ]

            for i in xrange(1, len(latest_commits)):
                deltas.append((weight(i), latest_commits[i-1].date - latest_commits[i].date))

            for weight, delta in deltas:
                log.debug('DELTA: %s, WEIGHT: %s' % (timedelta_to_string(delta), weight))

            interval_to_check = reduce(operator.add, (delta2secs(delta) * weight for weight, delta in deltas))

            interval_to_check = min(
                max(interval_to_check, CHECK_INTERVAL_MIN),
                CHECK_INTERVAL_MAX)
        else:
            interval_to_check = CHECK_INTERVAL_MIN

        interval_to_check = datetime.timedelta(0, interval_to_check)
        log.debug('next check will be after %s' % timedelta_to_string(interval_to_check))

        self.last_check_at = datetime.datetime.utcnow()
        self.next_check_at = self.last_check_at + interval_to_check
        self.save()

    @property
    def commits(self):
        return Commit.objects.find({'slug': self.slug})

    def update_last_access(self):
        self.last_access = datetime.datetime.utcnow()
        self.save()
        return ''



class Feed(Document):
    collection = 'feeds'

    def init(self, _id = None, ignore = [], watch = []):
        self._id = _id
        self.ignore = ignore
        self.watch = watch
        self.num_items = 0
        super(Feed, self).__init__(slug)


    def update(self):
        log = logging.getLogger('django_vcs_watch.feed.update')
        log.debug('updating %s' % self._id)

        if self.ignore is None:
            self.ignore = []

        if self.watch is None:
            self.watch = []

        if self.num_items is None:
            self.num_items = 0

        watch_query = ' || '.join(
            ' && '.join(
                "this.%s == '%s'" % item for item in rule.items())
                for rule in self.watch)

        ignore_query = ' || '.join(
            ' && '.join(
                "this.%s == '%s'" % item for item in rule.items())
                for rule in self.ignore)

        if not watch_query:
            return

        query = {'$where': '(%s) && !(%s)' % (watch_query, ignore_query or '0')}

        last_item = FeedItem.objects.find_one(dict(slug = self._id))

        if last_item is not None:
            query['date'] = {'$gt': last_item.date}

        Commit.objects.ensure_index([('slug', 1)])
        Commit.objects.ensure_index([('author', 1)])

        from pymongo.dbref import DBRef

        for commit in Commit.objects.find(query):
            commit_ref = DBRef(commit.objects.collection_name, commit._id)
            if FeedItem.objects.find_one(
                    dict(slug = self._id, commit = commit_ref)) is not None:
                logging.error('UPS I DID IT AGAIN!')

            FeedItem(slug = self._id, date = commit.date, commit = commit).save()
            self.num_items += 1

        self.save()


    def full_update(self):
        """ Drop all items and fill the feed with filtered items from scratch. """
        FeedItem.objects.remove(dict(slug = self._id))
        self.update()



class FeedItem(Document):
    collection = 'feed_items'

    class Meta:
        ordering = [('date', DESCENDING)]



def _init_mongo_connection(sender, **kwargs):
    Repository.objects.db = mongo()

from django.db.models.signals import class_prepared
class_prepared.connect(_init_mongo_connection)


def create_user_feed(instance, **kwargs):
    feed_id = get_user_feed_slug(instance)

    feed = Feed.objects.find_one(dict(_id = feed_id))

    if feed is None:
        feed = Feed(
            _id = feed_id,
            user_id = instance.id
        )
        feed.save()


from django.db.models import signals

signals.post_save.connect(create_user_feed, sender = User)



import re
import logging
import subprocess
from dateutil.parser import parse as parse_date
from dateutil.tz import tzutc
from xml.etree import ElementTree as ET

from django_vcs_watch.utils import \
    strip_timezone, \
    guess_encoding, \
    DiffProcessor

from django_vcs_watch.settings import \
    REVISION_LIMIT

def get_updates(url, last_rev, username = None, password = None):
    log = logging.getLogger('django_vcs_watch.backend.svn')

    command = ['svn', 'log', '--non-interactive']

    if last_rev:
        command += ['-r', 'HEAD:%s' % last_rev]

    command += ['--limit', '%s' % REVISION_LIMIT]

    if username:
        command += ['--username', username]

    if password:
        command += ['--password', password]

    command += ['--xml', url]

    log.debug(re.sub(r"--password '.*'", "--password 'xxxxx'", ' '.join(command)))

    svn = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    xml, stderr = svn.communicate()

    if svn.returncode != 0 or stderr or not xml:
        log.error('svn command failed with code %d and message %r' % (svn.returncode, stderr))
        raise Exception(' '.join(stderr.splitlines()))

    xml_e = ET.fromstring(xml)

    commits = []
    diff_processor = DiffProcessor()

    for entry_e in xml_e.findall('logentry'):
        revision = entry_e.attrib['revision']
        if revision == last_rev:
            continue

        author = entry_e.find('author').text
        msg = entry_e.find('msg').text
        date = parse_date(entry_e.find('date').text).astimezone(tzutc())

        command = ['svn', 'diff', '-c', str(revision), url]
        log.debug('fetching diff: %r' % ' '.join(command))

        svn = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        diff, stderr = svn.communicate()

        commits.insert(0, dict(
                           revision = revision,
                           changes = diff_processor.process(guess_encoding(diff)),
                           message = msg or u'',
                           date = strip_timezone(date),
                           author = author
                           )
                    )
    return commits




from django.conf import settings

__all__ = [
    'VCS_ONLY_PUBLIC_REPS',
    'VCS_URL_REWRITER',
    'CHECK_INTERVAL_MIN',
    'CHECK_INTERVAL_MAX',
    'REVISION_LIMIT',
    'DELAY_BETWEEN_UPDATE_CHECK',
    'MONGO_URL',
    'MONGO_DB',
]

VCS_ONLY_PUBLIC_REPS = getattr(settings, 'VCS_ONLY_PUBLIC_REPS', False)
VCS_URL_REWRITER = getattr(settings, 'VCS_URL_REWRITER', lambda x: x)

# Intervals in seconds
CHECK_INTERVAL_MIN = getattr(settings, 'VCS_CHECK_INTERVAL_MIN', 5 * 60)
CHECK_INTERVAL_MAX = getattr(settings, 'VCS_CHECK_INTERVAL_MAX', 5 * 60 * 60)

REVISION_LIMIT = getattr(settings, 'VCS_REVISION_LIMIT', 20)
DELAY_BETWEEN_UPDATE_CHECK = getattr(settings, 'VCS_DELAY_BETWEEN_UPDATE_CHECK', 30)

MONGO_URL = getattr(settings, 'MONGO_URL', 'localhost')
MONGO_DB = getattr(settings, 'MONGO_DB', 'vcs-watch')




import datetime
import pytz

from pdb import set_trace
from django import template
from django_vcs_watch.models import Commit, Repository, Feed, FeedItem
from django_vcs_watch.utils import get_user_feed_slug

register = template.Library()

@register.inclusion_tag('django_vcs_watch/top_repositories.html')
def top_repositories():
    results = Commit.objects \
        .group(
            keys = ['slug'],
            condition = {
                'date': {'$gte' : datetime.datetime.utcnow() - datetime.timedelta(7)}
            },
            initial = {'count': 0},
            reduce = 'function(obj, prev) { prev.count++; }'
        )
    results = sorted(results, key=lambda x: x['count'], reverse = True)[:10]
    results = Repository.objects \
        .find({'slug': {'$in': [r['slug'] for r in results]}})

    return {
        'objects': results
    }



def parse_token(token):
    tokens = token.split_contents()[1:]
    result = {}
    for token in tokens:
        name, value = token.split(':')
        if value[0] == value[-1] and value[0] in ('"', "'"):
            value = value[1:-1]
        else:
            value = template.Variable(value)
        result[name] = value
    return result



def resolve_variables(vars, context):
    result = vars.copy()
    for name, value in result.iteritems():
        resolve = getattr(value, 'resolve', None)
        if resolve is not None:
            result[name] = resolve(context)
    return result



class GetObjectList(template.Node):
    def __init__(self,
                 cls,
                 vars = {},
                 query = {},
                 on_page = 20,
                 map_func = lambda x: x,
                ):
        self.cls = cls
        self.vars = vars
        self.query = query
        self.on_page = int(vars.get('on_page', on_page))
        self.map_func = map_func


    def render(self, context):
        context['request'].GET.get('tb', None)

        vars = resolve_variables(self.vars, context)

        if callable(self.query):
            self.query = self.query(context, vars)

        cursor = self.cls.objects.find(self.query)

        context[self.vars.get('as', 'object_list')] = map(self.map_func, cursor[:self.on_page])
        return ''



class GetObject(template.Node):
    def __init__(self,
                 cls,
                 vars = {},
                 query = {},
                ):
        self.cls = cls
        self.vars = vars
        self.query = query


    def render(self, context):
        vars = resolve_variables(self.vars, context)

        if callable(self.query):
            self.query = self.query(context, vars)

        object = self.cls.objects.find_one(self.query)
        context[self.vars.get('as', 'object')] = object
        return ''



def get_tb(context):
    request = context['request']
    if 'tb' in request.GET:
        return pytz.utc.localize(
            datetime.datetime.utcfromtimestamp(
                int(request.GET.get('tb'))))
    return pytz.utc.localize(datetime.datetime.utcnow())



@register.tag
def get_all_commits(parser, token):
    vars = parse_token(token)
    return GetObjectList(
        cls = Commit,
        vars = vars,
        query = lambda context, vars: {
            'date' : {'$lt': get_tb(context)}
        },
    )


@register.tag
def get_commits_with_slug(parser, token):
    vars = parse_token(token)
    return GetObjectList(
        cls = Commit,
        vars = vars,
        query = lambda context, vars: dict(
            slug = vars['slug'],
            date = {'$lt': get_tb(context)}
        ),
    )

@register.tag
def get_commits_from_user_feed(parser, token):
    vars = parse_token(token)

    def process_commit(feed_item):
        commit = feed_item.commit
        commit.from_filtered_feed = True
        return commit

    return GetObjectList(
        cls = FeedItem,
        vars = vars,
        query = lambda context, vars: dict(
            slug = get_user_feed_slug(context['request'].user),
            date = {'$lt': get_tb(context)}
        ),
        map_func = process_commit,
    )

@register.tag
def get_commits_by(parser, token):
    vars = parse_token(token)
    return GetObjectList(
        cls = Commit,
        vars = vars,
        query = lambda context, vars: dict(
            author = vars['author'],
            date = {'$lt': get_tb(context)}
        ),
    )


@register.tag
def get_user_feed(parser, token):
    vars = parse_token(token)
    return GetObject(
        cls = Feed,
        vars = vars,
        query = lambda context, vars: dict(
            _id = get_user_feed_slug(context['request'].user),
        ),
    )




import re
import pymongo

from collections import defaultdict
from pdb import set_trace

from django.contrib.auth.decorators import login_required
from django.core.paginator import Paginator, InvalidPage
from django.http import HttpResponseForbidden, HttpResponseRedirect, HttpResponse, Http404
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext, loader
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST
from django.views.generic.create_update import apply_extra_context, redirect

from django_vcs_watch.models import Repository, Commit, Feed
from django_vcs_watch.forms import RepositoryForm
from django_vcs_watch.utils import get_user_feed_slug

_WRAPPERS = defaultdict(lambda: lambda x:x, {
    'repositories': Repository,
    'commits': Commit,
})


@login_required
def profile(request):
    from django.views.generic.list_detail import object_list
    return object_list(
            request,
            queryset=request.user.repository_set.all())


def autocomplete(request, **kwargs):
    q = request.GET.get('q', '')
    #db.repositories.ensure_index([('url', pymongo.ASCENDING)])

    return object_list(
            request,
            cls = Repository,
            query = {'url': re.compile(q)},
            **kwargs)

# TODO move this to separate django app like django-annoing
from django.contrib.syndication.views import feed as contrib_feed


def feed(request, slug, param = '', feed_dict = None):
    url = slug
    if param:
        url += '/' + param
    return contrib_feed(request, url, feed_dict = feed_dict)


def create(request,
           template_name,
           form_class = RepositoryForm,
           post_save_redirect = None,
           extra_context = {},
           template_loader = loader,
           context_processors = None,
          ):
    if request.method == 'POST':
        form = form_class(request.POST, request.FILES)
        if form.is_valid():
            new_object = form.save()
            return redirect(post_save_redirect, new_object)
    else:
        form = form_class()

    t = template_loader.get_template(template_name)
    c = RequestContext(request, {
        'form': form,
    }, context_processors)
    apply_extra_context(extra_context, c)
    return HttpResponse(t.render(c))


def object_detail(request,
                  cls,
                  query = {},
                  extra_context = {},
                  template_name = None,
                  template_name_field = None,
                  template_object_name = 'object',
                  template_loader = loader,
                  context_processors = None,
                  **kwargs
                 ):
    if callable(extra_context):
        extra_context = extra_context(request, **kwargs)

    if callable(query):
        query = query(request, **kwargs)

    object = cls.objects.find_one(query)

    if object is None:
        raise Http404, 'Object was not found in collection "%s"' % cls.collection.name()

    if not template_name:
        template_name = "%s_detail.html" % cls.objects.collection_name

    if template_name_field:
        template_name_list = [getattr(obj, template_name_field), template_name]
        t = template_loader.select_template(template_name_list)
    else:
        t = template_loader.get_template(template_name)

    c = RequestContext(request, {
            template_object_name: object,
        }, context_processors)
    apply_extra_context(extra_context, c)
    return HttpResponse(t.render(c))


def object_list(request,
                cls,
                query = {},
                paginate_by = None,
                page = None,
                allow_empty = True,
                template_name = None,
                template_loader = loader,
                extra_context = {},
                context_processors = None,
                template_object_name = 'object_list',
                mimetype = None,
                map_func = lambda x: x,
                **kwargs):
    """
    Generic list of objects.

    Templates: ``<collection_name>_list.html``
    Context:
        object_list
            list of objects
        is_paginated
            are the results paginated?
        results_per_page
            number of objects per page (if paginated)
        has_next
            is there a next page?
        has_previous
            is there a prev page?
        page
            the current page
        next
            the next page
        previous
            the previous page
        pages
            number of pages, total
        hits
            number of objects, total
        last_on_page
            the result number of the last of object in the
            object_list (1-indexed)
        first_on_page
            the result number of the first object in the
            object_list (1-indexed)
        page_range:
            A list of the page numbers (1-indexed).
    """
    if callable(extra_context):
        extra_context = extra_context(request, **kwargs)

    if callable(query):
        query = query(request, **kwargs)

    cursor = cls.objects.find(query)

    if paginate_by:
        paginator = Paginator(cursor, paginate_by, allow_empty_first_page=allow_empty)
        if not page:
            page = request.GET.get('page', 1)
        try:
            page_number = int(page)
        except ValueError:
            if page == 'last':
                page_number = paginator.num_pages
            else:
                # Page is not 'last', nor can it be converted to an int.
                raise Http404
        try:
            page_obj = paginator.page(page_number)
        except InvalidPage:
            raise Http404
        c = RequestContext(request, {
            template_object_name: map(map_func, page_obj.object_list),
            'paginator': paginator,
            'page_obj': page_obj,

            # Legacy template context stuff. New templates should use page_obj
            # to access this instead.
            'is_paginated': page_obj.has_other_pages(),
            'results_per_page': paginator.per_page,
            'has_next': page_obj.has_next(),
            'has_previous': page_obj.has_previous(),
            'page': page_obj.number,
            'next': page_obj.next_page_number(),
            'previous': page_obj.previous_page_number(),
            'first_on_page': page_obj.start_index(),
            'last_on_page': page_obj.end_index(),
            'pages': paginator.num_pages,
            'hits': paginator.count,
            'page_range': paginator.page_range,
        }, context_processors)
    else:
        c = RequestContext(request, {
            template_object_name: map(map_func, cursor),
            'paginator': None,
            'page_obj': None,
            'is_paginated': False,
        }, context_processors)
        if not allow_empty and len(cursor) == 0:
            raise Http404
    for key, value in extra_context.items():
        if callable(value):
            c[key] = value()
        else:
            c[key] = value
    if not template_name:
        template_name = "%s_list.html" % cls.objects.collection_name
    t = template_loader.get_template(template_name)
    return HttpResponse(t.render(c), mimetype=mimetype)



def commit(request, repository_slug, revision):
    mode = request.GET.get('mode', 'full')
    assert(mode in ('full', 'files'))

    file = request.GET.get('file', None)
    if file is not None:
        commit = Commit.objects.find_one(dict(
              slug = repository_slug,
              revision = revision)
        )
        if not commit:
            raise Http404, 'Object was not found in collection "%s"' % Commit.collection.name()

        obj = None
        for f in commit.changes.changed:
            if f.filename == file:
                obj = f
                break

        if obj is None:
            raise Http404, 'File %s was not found in commit "%s@%s"' % (file, commit.slug, commit.revision)

        return render_to_response(
          'django_vcs_watch/diff.html',
          dict(file = obj),
        )


    return object_detail(
            request,
            Commit,
            query = dict(slug = repository_slug, revision = revision),
            template_name = 'django_vcs_watch/commit_detail_%s.html' % mode,
        )


def get_user_feed(request):
    user = request.user
    feed_slug = get_user_feed_slug(user)
    return Feed.objects.find_one(dict(_id = feed_slug))


def refresh_feed(request, template_name = 'django_vcs_watch/refresh_feed.html'):
    feed = get_user_feed(request)
    feed.update()

    next = request.POST.get('next', None)
    if next is not None:
        return HttpResponseRedirect(next)

    return render_to_response(
        template_name,
        dict(feed = feed),
    )

def get_rule(request):
    rule_fields = ('author', 'slug')

    return dict(filter(
        lambda x: x[0] in rule_fields,
        request.POST.items()
    ))

@require_POST
def ignore(request):
    feed = get_user_feed(request)

    if feed.ignore is None:
        feed.ignore = []

    rule = get_rule(request)

    feed.ignore.append(rule)
    feed.save()

    if 'author' in rule:
        if 'slug' in rule:
            message = _('<div class="info">Now all commits from %(author)s in %(slug)s will be ignored. Boo-ga-ga!</div>')
        else:
            message = _('<div class="info">Now all commits from %(author)s will be ignored. Moo-ha-ha!</div>')
    else:
        message = _('<div class="error">Hm, it seems that we have no message for this case :).</div>')

    request.user.message_set.create(message = message % rule)

    next = request.POST.get('next', None)
    if not next:
        next = request.META.get('HTTP_REFERER', '/')

    return HttpResponseRedirect(next)


@require_POST
def unignore(request):
    feed = get_user_feed(request)

    if feed.ignore is None:
        feed.ignore = []

    rule = get_rule(request)

    if rule:
        feed.ignore = filter(lambda x: x != rule, feed.ignore)
    else:
        feed.ignore = []

    feed.save()

    if rule:
        request.user.message_set.create(message = _('Rule was removed.'))
    else:
        request.user.message_set.create(message = _('All rules were removed.'))

    next = request.POST.get('next', '/')
    return HttpResponseRedirect(next)


@require_POST
def watch(request):
    feed = get_user_feed(request)

    if feed.watch is None:
        feed.watch = []

    rule = get_rule(request)

    feed.watch.append(rule)
    feed.save()

    if 'slug' in rule:
        message = _('<div class="info">Now you watch on all commits to %(slug)s. Horay!</div>')
    elif 'author' in rule:
        message = _('<div class="info">Now you watch on all commits by %(author)s. Yee!</div>')
    else:
        message = _('<div class="error">Hm, it seems that we have no message for this case :).</div>')

    request.user.message_set.create(message = message % rule)

    next = request.POST.get('next', None)
    if not next:
        next = request.META.get('HTTP_REFERER', '/')

    return HttpResponseRedirect(next)


@require_POST
def unwatch(request):
    feed = get_user_feed(request)

    if feed.watch is None:
        feed.watch = []

    rule = get_rule(request)

    if rule:
        feed.watch = filter(lambda x: x != rule, feed.watch)
    else:
        feed.watch = []

    feed.save()

    if rule:
        request.user.message_set.create(message = _('Rule was removed.'))
    else:
        request.user.message_set.create(message = _('All rules were removed.'))

    next = request.POST.get('next', '/')
    return HttpResponseRedirect(next)




from django.contrib.syndication.feeds import FeedDoesNotExist, Feed
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import string_concat

from django_vcs_watch.models import Repository

class LatestRepositories(Feed):
    title = _('Last added repositories')
    description = _('Last added VCS repositories')

    def link(self):
        return reverse('vcs-watch-repositories')

    def items(self):
        return Repository.objects.filter(public=True).exclude(updated_at=None).order_by('-created_at')[:20]

    def item_pubdate(self, item):
        return item.created_at


class LatestCommits(Feed):
    def get_object(self, bits):
        if len(bits) != 1:
            raise ObjectDoesNotExist
        repository = Repository.objects.get(slug__exact=bits[0])
        repository.update_last_access()
        return repository

    def description(self, obj):
        if obj.last_error is not None:
            return string_concat(
                '<b>',
                _('We encounter an error during downloading diffs from this repository:'),
                '</b><br />',
                obj.last_error)

    def title(self, obj):
        return _('Latest changes at %s') % obj.url

    def link(self, obj):
        if not obj:
            raise FeedDoesNotExist
        return obj.get_absolute_url()

    def items(self, obj):
        return obj.revision_set.all()

    def item_pubdate(self, item):
        return item.date

    def item_author_name(self, item):
        return item.author




from setuptools import setup, find_packages

setup(
    name = 'django-vcs-watch',
    version = '0.2.1',
    description = 'RSS proxy for different VCS/DVCS.',
    keywords = 'django apps svn',
    license = 'New BSD License',
    author = 'Alexander Artemenko',
    author_email = 'svetlyak.40wt@gmail.com',
    url = 'http://github.com/svetlyak40wt/django-vcs-watch/',
    install_requires = ['django-globals', 'django-fields', 'python-dateutil'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Plugins',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    package_dir = {'': 'src'},
    packages = find_packages('src', exclude = ['example']),
    include_package_data = True,
    zip_safe = False,
)





import os
import logging
import time
import datetime

from itertools import chain
from django.conf import settings

from django_extensions.management.jobs import HourlyJob
from django_vcs_watch.models import Repository, Commit, Feed
from django_vcs_watch.settings import DELAY_BETWEEN_UPDATE_CHECK
from django_vcs_watch.utils import mongo

class Job(HourlyJob):
    help = "Update VCS feeds"

    def execute(self):
        log = logging.getLogger('django_vcs_watch.jobs.update_feeds')

        pidfile = getattr(settings, 'VCS_WATCH_PID_FILE', '/tmp/vcswatch_update_feeds.pid')

        if os.path.exists(pidfile):
            pid = int(open(pidfile).read())
            if os.path.exists('/proc/%d' % pid):
                log.warning('Job already running with pid %d (pidfile "%s" exists)' % (pid, pidfile))
                return

        f = open(pidfile, 'w')
        try:
            f.write(str(os.getpid()))
        finally:
            f.close()

        try:
            while True:
                if self.update_repositories():
                    self.update_feeds()
                time.sleep(DELAY_BETWEEN_UPDATE_CHECK)
        finally:
            os.remove(pidfile)


    def update_repositories(self):
        log = logging.getLogger('django_vcs_watch.jobs.update_repositories')

        new_reps = Repository.objects.find({'next_check_at': None})
        reps_to_update = Repository.objects.find({'next_check_at': {'$lte': datetime.datetime.utcnow()}})

        num_commits_before = Commit.objects.count()

        for rep in chain(new_reps, reps_to_update):
            log.debug('rep to update: %s' % rep.url)
            try:
                rep.update()
            except Exception, e:
                log.exception("can't update rep %s" % rep)

        Commit.objects.ensure_index([('date', 1)])

        return num_commits_before != Commit.objects.count()


    def update_feeds(self):
        log = logging.getLogger('django_vcs_watch.jobs.update_feeds')

        feeds = Feed.objects.find()

        for feed in feeds:
            log.debug('feed to update: %s' % feed._id)
            try:
                feed.update()
            except Exception, e:
                log.exception("can't update feed %s" % feed._id)

